# Copyright (c) 2013-2014  Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from .scanner import start, end, Scanner, Cursor
import collections

# This object is used for an unmatched Token when the Lexer is
# in sticky mode. Its just a value that serves as a placeholder
# and that is not compatible with any strings. This will make sure
# that you'll notice when you try to use an unmatched Token which
# was initialize with the sticky.
class sticky(object):

    def __init__(self, char):
        super(sticky, self).__init__()
        self.char = char

    def __repr__(self):
        return '<sticky %r>' % self.char

class Token(object):
    r""" This class represents a token from an input stream. A token
    has a value, type identifier and the cursor position it was found
    at. When the end of an input stream was reached, a Token with a
    None value is created by the :class:`Lexer`. Checking for the
    boolean value of the Token yields False in that special case.

    If the :class:`Lexer` is configured to also yield at characters
    that could not be gathered to a Token, the Token may contain a
    single-character string as its value only and the identifier is
    None.

    Additional attributes of a Token can be defined by passing keyword
    arguments to the constructor.

    As of version 0.5.1, a Token can also be "sticky", which means that
    its :attr:`value` is a :class:`sticky` object. This kind of Token
    is yielded when the :class:`Lexer` is in sticky mode and a
    character could not be identified. A sticky Token is also an
    unmatched Token.

    .. attribute:: id

        The identifier of the Token which can be any object that was
        defined to represent the type of the Token. Numbers or strings
        are very common.

        If this attribute is None, the Token is an un-identified Token
        which was not declared in any of the token type declarations or
        if the Token represents the end of the input stream in which
        case the :attr:`value` is None, too.

    .. attribute:: value

        The value of the Token, a string unless the Token marks the end
        of the input stream.

    .. attribute:: cursor

        The cursor position the Token was found at in the input stream.
        A :class:`scum.scanner.Cursor` instance.
    """

    def __init__(self, token_id, value, cursor, attrs={}):
        assert isinstance(cursor, Cursor)
        if not value and token_id is not None:
            raise RuntimeError('a Token without value can not have an ID')
        if isinstance(value, sticky) and token_id is not None:
            raise RuntimeError('a sticky Token must have an ID of None')

        super(Token, self).__init__()
        self.id = token_id
        self.value = value
        self.cursor = cursor.copy()
        self.attrs = dict(attrs)

    def __nonzero__(self):
        return bool(self.value)

    __bool__ = __nonzero__ # Python 3

    def __repr__(self):
        name = self.__class__.__name__
        if not self.value:
            return '<%s at end>' % name

        value = str(self.value)
        if len(value) > 20:
            value = value[:9] + '...' + value[9:]

        result = '<%s(%s) %r at %s' % (name, self.id, value, self.cursor)
        if self.attrs:
            result += ' %s' % self.attrs
        return result + '>'

    def __getattr__(self, name):
        try:
            return self.attrs[name]
        except KeyError:
            raise AttributeError('Token has not attribute %s' % name)

    @property
    def lineno(self):
        r""" Quick property to access the line-number of the Token
        from its :attr:`cursor` attribute. """

        return self.cursor.line

    @property
    def unmatched(self):
        r""" Returns True if this is an unmatched Token, meaning
        that there was not extractor that could identify this Token.
        Depending on the :attr:`Lexer.unmatched_mode`, the Token can
        contain a single character or a multi-character string. """

        return self.value and self.id is None

    @property
    def sticky(self):
        r""" Returns True if this is a sticky Token. A sticky Token
        will always also be an :attr:`unmatched` Token. """

        return isinstance(self.value, sticky)

    @property
    def invalid(self):
        r""" Returns True if this is an invalid Token which represents
        the end of the input stream. """

        return not self.value and self.id is None

class TokenExtractor(object):
    r""" The base class for objects that are capable of extracting
    a :class:`Token` from a :class:`Scanner`. """

    def extract_token(self, scanner, token_id, cursor, token_attrs):
        r""" Called to extract and return a :class:`Token` from
        the *scanner*. Return None to tell that no Token could be
        extracted. """

        raise NotImplementedError

class TokenExtractorWrapper(TokenExtractor):
    r""" A wrappr for callable objects that accept only a Scanner
    as their first and only argument. """

    def __init__(self, func):
        super(TokenExtractorWrapper, self).__init__()
        self.func = func

    def extract_token(self, scanner, token_id, cursor, token_attrs):
        result = self.func(scanner)
        if result:
            return Token(token_id, result, cursor, token_attrs)
        else:
            return None

class Lexer(object):
    r""" The Lexer uses the :class:`Scanner` class to analyze an input
    character stream and split it into Tokens which are declared via
    :meth:`declare_token`.

    A Lexer is re-usable as the Scanner can be plugged in and out. """

    def __init__(self, unmatched_mode='no-join'):
        super(Lexer, self).__init__()
        self.__tokendefs = []
        self.__token = None
        self.__reserved_token = None
        self.__scanner = None
        self.unmatched_mode = unmatched_mode

    def __iter__(self):
        if not self.__scanner:
            raise RuntimeError('Lexer has no source assigned')

        return TokenStream(self)

    @property
    def token(self):
        r""" Returns the last Token that was extracted from the input
        stream. The ID of the Token is None if no Token declaration
        was able to process the value of the Token (which is a one-
        character string). If the value of the Token is None, the end
        of the input stream was reached. """

        return self.__token

    @property
    def unmatched_mode(self):
        r""" This property defines whether the Lexer joins unmatched
        characters into one Token or yields each unmatched character
        with its own Token instance. This propery is represented by
        a string which can have three different modes:

        * ``'no-join'``
        * ``'join'``
        * ``'join-to-newline'``
        * ``'sticky'``: When this mode is set, the Lexer will stick
        to the current position in the Scanner until it was manually
        moved. An unmatched Token will be returned where the
        :attr:`Token.value` is the :data:`sticky` object. __Note__ that
        using the ``'sticky'`` mode __requires__ direct interaction
        with the Lexers underlying Scanner, otherwise you'll get the
        same sticky Token all over again.
        """

        return self.__unmatched_mode

    @unmatched_mode.setter
    def unmatched_mode(self, mode):
        if mode not in ('no-join', 'join', 'join-to-newline', 'sticky'):
            raise ValueError('invalid unmatched_mode %r' % mode)
        self.__unmatched_mode = mode

    @property
    def is_initialized(self):
        r""" Returns True if the Lexer is initialized and lexing is
        in progress (that is, when a Scanner was set as the source of
        the Lexer via :meth:`set_source`). """

        return bool(self.__scanner)

    @property
    def source(self):
        r""" Returns the source of the Lexer, which is None when the
        Lexer is uninitialized or a :class:`Scanner` instance. """

        return self.__scanner

    def declare_token(self, token_id, extractor, **token_attrs):
        r""" Declares a Token to the Lexer. The order of declaration
        defines the order they are processed, the first declared Token
        is applied first. The first matching declaration is used as the
        actual Token. If *skip* is True, the Token will be skipped by
        the Lexer and not even yielded but instead the next Token is
        read.

        Two declarations with the same ID are possible in case that
        is of any use for someone.

        Passing None for the *token_id* will cause the Lexer to
        skip the Token completely when it was extracted and instead
        extract the next Token.

        The *extractor* argument must be either a callable object or an
        instance of the :class:`TokenExtractor` class. If it is a
        callable object, it must accept a :class:`Scanner` as its only
        argument and return a string of the extracted Token value.
        Both, a callable object or a TokenExtractor, do not need to
        take care of re-setting the Scanners cursor to the original
        position.

        For extended usage, Token attributes can be specified by
        passing additional keyword arguments which will be set as
        attributes for the Token generated by this extractor.

        :raise RuntimeError: If the Lexer has a source already
            assigned (see :attr:`is_initialized`). """

        if self.__scanner:
            raise RuntimeError('can not declare new tokens while lexing is '
                               'in progress')

        if not isinstance(extractor, TokenExtractor):
            if isinstance(extractor, collections.Callable):
                extractor = TokenExtractorWrapper(extractor)
            else:
                raise TypeError('extractor is not callable nor TokenExtractor')

        self.__tokendefs.append((token_id, extractor, token_attrs))

    def set_source(self, scanner):
        r""" Set the source of the input stream to a :class:`Scanner`
        instance. Raises TypeError if *scanner* is not such an instance.

        After this method was successfully called,
        :meth:`declare_token` can not be called until the Lexer
        was de-initialized with :meth:`unplug`.

        A RuntimeError is raised when this method is called multiple
        times without de-initializing the Lexer beforehand. """

        if self.__scanner:
            raise RuntimeError('can not overwrite existing source of lexer')
        if not isinstance(scanner, Scanner):
            raise TypeError('scanner must be Scanner instance')

        self.__scanner = scanner
        if scanner.char == start:
            scanner.next()

    def unplug(self):
        r""" Unplugs the source of the Lexer and de-initializes it.
        Raises a RuntimeError when called in an un-initialized state. """

        if not self.__scanner:
            raise RuntimeError('Lexer is already un-initialized')

        self.__scanner = None
        self.__token = None
        self.__reserved_token = None

    # backwards compatibility for v0.5.0
    unplug_source = unplug

    def __nexttoken(self):
        r""" Private method called by :meth:`next`. It simply tries
        to match all declared token extractors and returns the token
        that matched, or an unmatched Token when none matched. """

        scanner = self.__scanner
        cursor = scanner.cursor

        # Check if we are at the end of the input stream and yield
        # the invalid Token in that special case.
        if scanner.char == end:
            return Token(None, None, cursor)

        token = None

        # Try all declared Token matchers.
        for token_id, extractor, token_attrs in self.__tokendefs:
            result = extractor.extract_token(
                    scanner, token_id, cursor, token_attrs)

            # If an object was returned, it must be a Token instance.
            if result and not isinstance(result, Token):
                message = 'token declaration %r returned non-Token instance %s'
                raise TypeError(message % (token_id, result.__class__.__name__))
            elif result:
                token = result
                break

            # Re-set the Scanners cursor to its original position.
            scanner.cursor = cursor

        return token

    def next(self):
        r""" Reads the next Token from the input stream. The last
        extracted Token can be accessed via the :attr:`token`
        attribute.

        Returns a :class:`Token` object. An unmatched Token is returned
        when no Token could be extracted from the current position of
        the source Scanner. Depending on the :attr:`unmatched_mode`,
        this Token may contain a single character or a full string of
        all the characters that could not be extracted in a row.

        If a Token is extracted which was declared with a Token ID of
        the None value, the Token is skipped and the next Token is
        being extracted and returned.

        If the :attr:`Token.value` of the returned Token is None, the
        end of the input stream was reached.

        .. seealso:: :attr:`Token.unmatched`, :attr:`Token.invalid`

        .. note:: If an unmatched Token is being returned from this
                method, it is possible that the Lexer has the next
                Token already extracted and the underlying Scanner
                might not be located at the expected position. This
                is not the case in sticky mode.

        Raises a RuntimeError if the Lexer is not initialized. """

        scanner = self.__scanner
        if not scanner:
            raise RuntimeError('Lexer has no source assigned')

        if self.__reserved_token:
            token, self.__reserved_token = (self.__reserved_token, None)
            self.__token = token
            return token

        token = self.__nexttoken()

        # If no Token could be extracted, return an unmatched Token.
        # Gather additional unmatched characters depending on the
        # value of `unmatched_mode`.
        if token is None:

            cursor = scanner.cursor
            mode = self.__unmatched_mode
            if mode == 'sticky':
                value = sticky(scanner.char)
            else:
                value = scanner.get_and_next()


            # Gather additional characters that we can not match to
            # add them to the unmatched Token we'll be returning later.
            if mode in ('join', 'join-to-newline'):
                # Terminate at this character.
                term = end if mode == 'join' else scanner.newline_rep

                while scanner.char != term:
                    # Try to extract the next Token. If there is one,
                    # we will put it into the reserved place. If it
                    # could not be extracted, the next character is
                    # added to the unmatched Token value.
                    token = self.__nexttoken()
                    if token is not None:
                        self.__reserved_token = token
                        break

                    value += scanner.get_and_next()

                # The new-line should be included in the join-to-newline
                # mode of the unmatched_mode member. This should only
                # be done if we are still right behind the last character
                # we've added to `value` (which is the case if we did not
                # stop because a real token was extracted).
                if token is None and scanner.char == scanner.newline_rep:
                    value += scanner.get_and_next()

            # Create an unmatched Token.
            token = Token(None, value, cursor)

        # If the token we get here is unmatched, it was extracted
        # by a token extractor with a Token ID of None, which we
        # will not yield. Just grab the next Token.
        elif token.unmatched:
            token = self.next()

        self.__token = token
        return token

    def trynext(self, *token_ids):
        r""" This method is similar to :meth:`next` but will only try
        extract a Token of one of the IDs specified with *\*token_ids*.
        If no Token can be extracted, None is returned and the source
        is reverted to its original position. This function will never
        return an unmatched or invalid Token.

        Calling this function will not mess up Token extraction
        precedence as the Tokens are extracted as usual. One can assume
        that the Scanner is right where the last Token ended and no
        Token is reserved (if the reserved Token does not match one of
        the requested IDs, the reversed Token will be unreserved).

        *New in v0.5.1* """

        if not self.__scanner:
            raise RuntimeError('Lexer has no source assigned')

        if self.__reserved_token is not None:
            token = self.__reserved_token
            if token.id is None or token.id not in token_ids:
                return None
            else:
                self.__reserved_token = None
                self.__token = token
                return token

        self.__reserved_token = self.__nexttoken()
        if self.__reserved_token is None:
            # No token was extracted at all.
            return None

        return self.trynext(*token_ids)

    def unreserve(self):
        r""" It can sometimes happen that a Token is already reserved
        and ready to be returned on the next call to :meth:`next`. This
        can lead to unexpected behavior when one is actually expecting
        the Lexers source to be pointing at the location right where
        the current :attr:`token` ended, but it actually is positioned
        at the end of the reserved Token.

        If a Token is currently reserved, this method will remove the
        reserved Token and move the Scanner to the respective position.
        If no Token is reserved, this function will do nothing.

        *New in v0.5.1* """

        if self.__reserved_token is not None:
            self.__scanner.cursor = self.__reserved_token.cursor
            self.__reserved_token = None

    def get_and_next(self):
        r""" Returns the current Token and extracts the next. """

        try:
            return self.__token
        finally:
            self.next()

    def token_proxy(self):
        r""" Requires the :mod:`werkzeug` module. Returns a LocalProxy
        object for the Token in the lexer. """

        from werkzeug.local import LocalProxy
        return LocalProxy(lambda: self.token)

class TokenStream(object):
    r""" This class is an iterator for a :class:`Lexer` object
    and basically represents a stream of :class:`Tokens` that are
    yielded by the Lexer. """

    def __init__(self, lexer):
        if not lexer.is_initialized:
            raise ValueError('lexer is not initialized')
        super(TokenStream, self).__init__()
        self.__lexer = lexer

    def __iter__(self):
        return self

    def next(self):
        token = self.__lexer.next()
        if token.value is None:
            raise StopIteration
        return token

    __next__ = next

